@@ -84,7 +84,7 @@ int
 setinterface(int socket_fd, const char *interface_name)
 {
     struct ifreq interface;
-    memset(&interface, 0, sizeof(interface));
+    memset(&interface, 0, sizeof(struct ifreq));
     strncpy(interface.ifr_name, interface_name, IFNAMSIZ);
     int res = setsockopt(socket_fd, SOL_SOCKET, SO_BINDTODEVICE, &interface,
                          sizeof(struct ifreq));
@@ -99,7 +99,7 @@ bind_to_address(int socket_fd, const char *host)
     if (host != NULL) {
         struct cork_ip ip;
         struct sockaddr_storage storage;
-        memset(&storage, 0, sizeof(storage));
+        memset(&storage, 0, sizeof(struct sockaddr_storage));
         if (cork_ip_init(&ip, host) != -1) {
             if (ip.version == 4) {
                 struct sockaddr_in *addr = (struct sockaddr_in *)&storage;
@@ -118,7 +118,9 @@ bind_to_address(int socket_fd, const char *host)
 }
 
 ssize_t
-get_sockaddr(char *host, char *port, struct sockaddr_storage *storage, int block)
+get_sockaddr(char *host, char *port,
+             struct sockaddr_storage *storage, int block,
+             int ipv6first)
 {
     struct cork_ip ip;
     if (cork_ip_init(&ip, host) != -1) {
@@ -167,18 +169,24 @@ get_sockaddr(char *host, char *port, struct sockaddr_storage *storage, int block
             return -1;
         }
 
+        int prefer_af = ipv6first ? AF_INET6 : AF_INET;
         for (rp = result; rp != NULL; rp = rp->ai_next)
-            if (rp->ai_family == AF_INET) {
-                memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in));
+            if (rp->ai_family == prefer_af) {
+                if (rp->ai_family == AF_INET)
+                    memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in));
+                else if (rp->ai_family == AF_INET6)
+                    memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in6));
                 break;
             }
 
         if (rp == NULL) {
-            for (rp = result; rp != NULL; rp = rp->ai_next)
-                if (rp->ai_family == AF_INET6) {
+            for (rp = result; rp != NULL; rp = rp->ai_next) {
+                if (rp->ai_family == AF_INET)
+                    memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in));
+                else if (rp->ai_family == AF_INET6)
                     memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in6));
-                    break;
-                }
+                break;
+            }
         }
 
         if (rp == NULL) {
@@ -299,4 +307,3 @@ validate_hostname(const char *hostname, const int hostname_len)
 
     return 1;
 }
-
