@@ -55,13 +55,19 @@
 #endif
 #endif
 
+#ifndef UPDATE_INTERVAL
+#define UPDATE_INTERVAL 10
+#endif
+
 /** byte size of ip4 address */
 #define INET_SIZE 4
 /** byte size of ip6 address */
 #define INET6_SIZE 16
 
 size_t get_sockaddr_len(struct sockaddr *addr);
-ssize_t get_sockaddr(char *host, char *port, struct sockaddr_storage *storage, int block);
+ssize_t get_sockaddr(char *host, char *port,
+                     struct sockaddr_storage *storage, int block,
+                     int ipv6first);
 int set_reuseport(int socket);
 
 #ifdef SET_INTERFACE
